from nautilus import Prior
from scipy.stats import norm
import numpy as np

class priorHandler(object):
    """ Contains functions for translating BOGfit priors into 
    Nautlius priors and translating Nautilus samples into model
    paremeters in each filter.

    Parameters
    ----------

    user_prior : dict
        The user specified prior which set out the priors for the model parameters
        and passes information about whether to let these vary for each filter or
        whether they follow an order-specified polynomial relationship.

    filter_names : list - str
        A list of filter names e.g., [F356W, F444W, ...]

    wavs : array
        A 1D array of effective wavelengths corresponding to each filter.
    
    """


    def __init__(self, user_prior, filter_names, wavs):
        self.user_prior = user_prior
        self.filter_names = filter_names
        self.wavs = wavs


    def translate(self):
        """ Translate a BOGfit prior into a Nautilus prior."""
        # Initiate Nautlius prior
        prior = Prior()

        # Loop over model parameters
        for key in self.user_prior.keys():
            param_prior_traits = self.user_prior[key]
            param_prior_dist = param_prior_traits[0]
            param_fit_type = param_prior_traits[1]

            # If user specifies 'Individual', add a fitter free-parameter for each filter.
            if param_fit_type == "Individual":
                for i in range(len(self.wavs)):
                    prior.add_parameter(f'{key}_{self.filter_names[i]}', dist=param_prior_dist)
            
            # If user specifies 'Polynomial', add a fitter free-parameter for each coefficient. 
            # e.g., For order 0, only one free-parameter is included for the whole fitting run a.k.a constant between filters.
            # e.g., For order 1, two free-parameter are included for the whole fitting run a.k.a straight-line relationship  with effective wavelength.
            # The lowest wavelength is chosen as the origin.
            elif param_fit_type == "Polynomial":
                prior.add_parameter(key+f'_C0', dist=param_prior_dist)
                poly_order = param_prior_traits[2]
                higher_order_dist = norm(loc=0, scale=1) # Priors for gradients, higher-order coefficients.
                for i in range(1,poly_order+1):
                    prior.add_parameter(key+f'_C{i}', dist=higher_order_dist)

            else:
                raise Exception("Fitting type not supported, please choose either 'Individual' or 'Polynomial'.")

        # Add free-parameters for a constant RMS background noise and constant Sersic poisson uncertainty scaling.
        prior.add_parameter(f'rms', dist=(0,1))
        prior.add_parameter(f'rms_sersic', dist=(0,100))
        
        return prior


    def revert(self, param_dict, wavs):
        """ Translate a Nautilus sample into a model parameters for each filter."""

        # Empty parameter array
        params_final = np.zeros((len(self.wavs), len(self.user_prior.keys())))

        # Loop over model parameters
        keys = list(self.user_prior.keys())
        for i in range(len(keys)):
            param_prior_traits = self.user_prior[keys[i]]
            param_fit_type = param_prior_traits[1]

            # If individual, add the Nautilus sample for each filter
            if param_fit_type == "Individual":
                for j in range(len(self.wavs)):
                    params_final[j,i] = param_dict[f"{keys[i]}_{self.filter_names[j]}"]
            
            # If polynomial, calculate the expected parameter in each filter given its effective wavlength
            elif param_fit_type == "Polynomial":
                poly_order = param_prior_traits[2]
                coeffs = np.zeros(poly_order+1)
                for k in range(poly_order+1):
                    coeffs[k] = param_dict[f"{keys[i]}_C{k}"]
                polywavs = np.power(np.tile(wavs-wavs[0], (poly_order+1,1)).T, np.arange(poly_order+1))
                comps = coeffs * polywavs
                comps_summed = np.sum(comps, axis=1)
                params_final[:,i] =comps_summed        

        return params_final



import numpy as np


class filter_set(object):
    """ Class for loading and manipulating sets of filter curves. This
    is where integration over filter curves to get photometry happens.

    Parameters
    ----------

    filt_list : list
        List of strings containing paths from the working directory to
        files where filter curves are stored. The filter curve files
        should contain an array of wavelengths in Angstroms followed by
        a column of relative transmission values.
    """

    def __init__(self, filt_list):
        self.filt_list = filt_list
        self.wavelengths = None
        self._load_filter_curves()
        self._calculate_min_max_wavelengths()
        self._calculate_effective_wavelengths()

    def _load_filter_curves(self):
        """ Loads filter files for the specified filt_list and truncates
        any zeros from either of their edges. """

        self.filt_dict = {}

        for filt in self.filt_list:
                self.filt_dict[filt] = np.loadtxt(filt, usecols=(0, 1))


    def _calculate_min_max_wavelengths(self):
        """ Finds the min and max wavelength values across all of the
        filter curves. """

        self.min_phot_wav = 9.9*10**99
        self.max_phot_wav = 0.

        for filt in self.filt_list:
            min_wav = (self.filt_dict[filt][0, 0]
                       - 2*(self.filt_dict[filt][1, 0]
                       - self.filt_dict[filt][0, 0]))

            max_wav = (self.filt_dict[filt][-1, 0]
                       + 2*(self.filt_dict[filt][-1, 0]
                       - self.filt_dict[filt][-2, 0]))

            if min_wav < self.min_phot_wav:
                self.min_phot_wav = min_wav

            if max_wav > self.max_phot_wav:
                self.max_phot_wav = max_wav

    def _calculate_effective_wavelengths(self):
        """ Calculates effective wavelengths for each filter curve. """

        self.eff_wavs = np.zeros(len(self.filt_list))

        for i in range(len(self.filt_list)):
            filt = self.filt_list[i]
            dlambda = self.make_bins(self.filt_dict[filt][:, 0])[1]
            filt_weights = dlambda*self.filt_dict[filt][:, 1]
            self.eff_wavs[i] = np.sqrt(np.sum(filt_weights*self.filt_dict[filt][:, 0])
                                       / np.sum(filt_weights
                                       / self.filt_dict[filt][:, 0]))
            


    def make_bins(self, midpoints, make_rhs=False):
        """ A general function for turning an array of bin midpoints into an
        array of bin left hand side positions and bin widths. Splits the
        distance between bin midpoints equally in linear space.

        Parameters
        ----------

        midpoints : numpy.ndarray
            Array of bin midpoint positions

        make_rhs : bool
            Whether to add the position of the right hand side of the final
            bin to bin_lhs, defaults to false.
        """

        bin_widths = np.zeros_like(midpoints)
        if make_rhs:
            bin_lhs = np.zeros(midpoints.shape[0]+1)
            bin_lhs[0] = midpoints[0] - (midpoints[1]-midpoints[0])/2
            bin_widths[-1] = (midpoints[-1] - midpoints[-2])
            bin_lhs[-1] = midpoints[-1] + (midpoints[-1]-midpoints[-2])/2
            bin_lhs[1:-1] = (midpoints[1:] + midpoints[:-1])/2
            bin_widths[:-1] = bin_lhs[1:-1]-bin_lhs[:-2]

        else:
            bin_lhs = np.zeros_like(midpoints)
            bin_lhs[0] = midpoints[0] - (midpoints[1]-midpoints[0])/2
            bin_widths[-1] = (midpoints[-1] - midpoints[-2])
            bin_lhs[1:] = (midpoints[1:] + midpoints[:-1])/2
            bin_widths[:-1] = bin_lhs[1:]-bin_lhs[:-1]

        return bin_lhs, bin_widths
    

    def resample_filter_curves(self, wavelengths):
        """ Resamples the filter curves onto a new set of wavelengths
        and creates a 2D array of filter curves on this sampling. """

        self.wavelengths = wavelengths  # Wavelengths for new sampling

        # Array containing filter profiles on new wavelength sampling
        self.filt_array = np.zeros((wavelengths.shape[0], len(self.filt_list)))

        # Array containing the width in wavelength space for each point
        self.widths = self.make_bins(wavelengths)[1]

        for i in range(len(self.filt_list)):
            filt = self.filt_list[i]
            self.filt_array[:, i] = np.interp(wavelengths,
                                              self.filt_dict[filt][:, 0],
                                              self.filt_dict[filt][:, 1],
                                              left=0, right=0)
            
    def get_photometry(self, spectrum, redshift, unit_conv=None):
        """ Calculates photometric fluxes. The filters are first re-
        sampled onto the same wavelength grid with transmission values
        blueshifted by (1+z). This is followed by an integration over
        the observed spectrum in the rest frame:

        flux = integrate[(f_lambda*lambda*T(lambda*(1+z))*dlambda)]
        norm = integrate[(lambda*T(lambda*(1+z))*dlambda))]
        photometry = flux/norm

        lambda:            rest-frame wavelength array
        f_lambda:          observed spectrum
        T(lambda*(1+z)):   transmission of blueshifted filters
        dlambda:           width of each wavelength bin

        The integrals over all filters are done in one array operation
        to improve the speed of the code.
        """

        if self.wavelengths is None:
            raise ValueError("Please use resample_filter_curves method to set"
                            + " wavelengths before calculating photometry.")

        redshifted_wavs = self.wavelengths*(1. + redshift)

        # Array containing blueshifted filter curves
        filters_z = np.zeros_like(self.filt_array)

        # blueshift filter curves to sample right bit of rest frame spec
        for i in range(len(self.filt_list)):
            filters_z[:, i] = np.interp(redshifted_wavs, self.wavelengths,
                                        self.filt_array[:, i],
                                        left=0, right=0)

        # Calculate numerator of expression
        flux = np.expand_dims(spectrum*self.widths*self.wavelengths, axis=1)
        flux = np.sum(flux*filters_z, axis=0)

        # Calculate denominator of expression
        norm = filters_z*np.expand_dims(self.widths*self.wavelengths, axis=1)
        norm = np.sum(norm, axis=0)

        photometry = np.squeeze(flux/norm)

        # This is a little dodgy as pointed out by Ivo, it should depend
        # on the spectral shape however only currently used for UVJ mags
        if unit_conv == "cgs_to_mujy":
            photometry /= (10**-29*2.9979*10**18/self.eff_wavs**2)

        return photometry


import numpy as np
import matplotlib.pyplot as plt
import petrofit as pf
from astropy.convolution.utils import discretize_model
import corner
from nautilus import Sampler
import time
import os
from astropy.modeling import models
import pandas as pd

from .prior_handler import priorHandler
from ..plotting import Plotter

from ..utils import filter_set

from tqdm import tqdm


dir_path = os.getcwd()
if not os.path.isdir(dir_path + "/mimical"):
    os.system('mkdir ' + dir_path + "/mimical")
    os.system('mkdir ' + dir_path + "/mimical/plots")
    os.system('mkdir ' + dir_path + "/mimical/posteriors")
    os.system('mkdir ' + dir_path + "/mimical/cats")



class mimical(object):
    """ Bayesian Observer of Galaxies - Fit.
    Fits multi-filter images of galaxies simultaeneously using Petrofit and Nautilus
    by specifying either an individual filter dependency for model parameters or
    a user-specified order polynomial. By default a Sersic profile is used however any
    Astropy model can be specified.

    Parameters
    ----------

    id : str
        An ID for the fitting run. Only really used for output files.

    images : array
        A 3D array of image data with slices for each filter. Each image
        must be the same shape.

    filt_list : str or list
        A list of path strings to the filter transmission curve files, relative
        to the current working directory. Must be in ascending order with effective wavelength.

    psfs : array
        A 3D array of normalised PSF images with slices for each filter. Each PSF image
        must be the same shape, and the same shape as the data images.

    astropy_model : array
        Astropy Fittable2DModel used to model the image data. The subsequent prior must include
        only and all parameters in the astropy_model.parameters variable, as well as a 'psf_pa' parameter.

    user_prior : dict
        The user specified prior which set out the priors for the model parameters
        and passes information about whether to let these vary for each filter or
        whether they follow an order-specified polynomial relationship.
    """


    def __init__(self, id, images, filt_list, psfs, user_prior, astropy_model=models.Sersic2D(), pool=None, timeout=3600):
        
        # Helper when only one image is passed
        if len(images.shape)==2:
            images = np.array(([images]))
        if (type(filt_list).__name__=='str') | (type(filt_list).__name__=='str_'):
            filt_list = [filt_list]
        if len(psfs.shape)==2:
            psfs = np.array(([psfs]))

        self.id = id

        print(f"Fitting object {id}")
        self.images = images
        self.psfs = psfs
        self.user_prior = user_prior
        self.astropy_model = astropy_model

        # Using the filter files, find the name of the filters and the effective wavelengths.
        self.filter_names = [x.split('/')[-1] for x in filt_list]
        self.wavs = filter_set([dir_path+'/'+x for x in filt_list]).eff_wavs / 1e4

        # Initiate the prior handler object, used to parse and translate priors and parameters.
        self.prior_handler = priorHandler(user_prior, self.filter_names, self.wavs)

        # Translate user specified prior into a prior parseable by nautlius.
        self.fitter_prior = self.prior_handler.translate()
        print(f"Fitting with parameters: {self.fitter_prior.keys}")

        self.t0 = time.time()
        self.calls = 0

        self.timeout=timeout
        self.pool = pool





    def lnlike(self, param_dict):
        """ Returns the log-likelihood for a given parameter vector. """

        # Translate parameter vector into model parameters in each filter.
        pars = self.prior_handler.revert(param_dict, self.wavs)

        # Define empty arrays for models and rms images.
        models = np.zeros_like(self.images)
        rms = np.zeros_like(self.images)
    
        # Loop over filters
        for i in range(len(self.wavs)):
            # Update the model and evaluate over a pixel grid.
            self.convolved_models[i].parameters = pars[i]
            model = discretize_model(model=self.convolved_models[i], 
                                     x_range=[0,self.images[i].shape[1]], 
                                     y_range=[0,self.images[i].shape[0]], 
                                     mode='center')

            # If, for whatever reason, the model has NaNs, set to zero and blow up errors.
            if np.isnan(np.sum(model)):
                models[i] = np.zeros_like(model)
                rms[i] = np.zeros_like(model) + 1e99

            # Else, append to respective arrays.
            else:
                models[i] = model
                rms[i] = param_dict['rms'] + (param_dict['rms_sersic']*np.sqrt(np.abs(model)))
                #rms[i] += 1.483 * median_abs_deviation(self.images[i].flatten())


        # Broadcast the 3D data and model arrays and sum through the resulting 3D log-likelihood array.
        log_like_array = np.log((1/(np.sqrt(2*np.pi*(rms**2))))) + ((-(self.images - models)**2) / (2*(rms**2)))
        log_like = np.sum(log_like_array.flatten())

        # Print calls
        self.calls+=1
        if not self.calls % 100:
            #print('Time % 100: ' + str(time.time()-self.t0))
            self.t0 = time.time()

        return(log_like)




    def fit(self):
        """ Runs the Nautlius sampler to fit models, and processes its output. """

        # Define empty models for each filter
        sersic_model = self.astropy_model
        self.convolved_models = []
        for i in range(len(self.wavs)):
            self.convolved_models.append(pf.PSFConvolvedModel2D(sersic_model, psf=self.psfs[i], oversample=(self.images.shape[2]/2, self.images.shape[1]/2, 15, 10)))

        # Check that the user specified prior contains the same parameters as the user specified model.
        if list(self.convolved_models[0].param_names) != list(self.user_prior.keys()):
            raise Exception("Prior labels do not match model parameters.")
        

        if os.path.isfile(dir_path+'/mimical/posteriors' + f'/{self.id}.txt'):
            self.success = True
            self.samples = pd.read_csv(dir_path+'/mimical/posteriors' + f'/{self.id}.txt', delimiter=' ').to_numpy()
            fit_dic = dict(zip((np.array((list(self.fitter_prior.keys)))+"_50").tolist(), np.median(self.samples, axis=0).tolist()))
            print(f"Loading existing posterior at " + dir_path + '/mimical/posteriors' + f'/{self.id}.txt')
            self.save_cat()
            print(" ")
            return fit_dic


        # Run sampler
        t0 = time.time()
        #sampler = Sampler(self.fitter_prior, self.lnlike, n_live=400, filepath= dir_path+'/bogout'+f'/{self.id}.hdf5', resume=True, pool=self.pool)
        sampler = Sampler(self.fitter_prior, self.lnlike, n_live=400, pool=self.pool)
        self.success = sampler.run(verbose=True, timeout=self.timeout)
        print(f"Sampling time (minutes): {(time.time()-t0)/60}")

        if self.success != True:

            print("Sampling failed (timeout).")
            print(" ")
            return {}
    
        else:
            # Sample the posterior information
            points, log_w, log_l = sampler.posterior()

            # Plot and save the corner plot
            corner.corner(points, weights=np.exp(log_w), bins=20, labels=np.array(self.fitter_prior.keys), color='purple', plot_datapoints=False, range=np.repeat(0.999, len(self.fitter_prior.keys)))
            plt.savefig(dir_path+'/mimical/plots' + f'/{self.id}_corner.pdf', bbox_inches='tight')

            # Sample an appropriately weighted posterior for representative samples.
            n_post = 10000
            indices = np.random.choice(np.arange(points.shape[0]), size = n_post, p=np.exp(log_w))
            self.samples = points[indices]
            samples_df = pd.DataFrame(data=self.samples, columns=self.fitter_prior.keys)
            samples_df.to_csv(dir_path+'/mimical/posteriors' + f'/{self.id}.txt', sep=' ', index=False)

            # Return the median-parameter model
            fit_dic = dict(zip((np.array((list(self.fitter_prior.keys)))+"_50").tolist(), np.median(self.samples, axis=0).tolist()))

            print("Sampling finished successfully.")
            print(" ")

            self.save_cat()

            return fit_dic
        
    
    def save_cat(self):

        user_samples = np.zeros((self.samples.shape[0], len(self.wavs), len(self.user_prior.keys())))
        # Get median Nautilus parameters and transalte into median model parameters.
        for j in tqdm(range(self.samples.shape[0])):
            # Get median Nautilus parameters and transalte into median model parameters.
            param_dict = dict(zip(list(self.fitter_prior.keys), self.samples[j]))
            pars = self.prior_handler.revert(param_dict, self.wavs)
            user_samples[j] = pars

        quantiles = np.percentile(user_samples, q=(16, 50, 84), axis=0)

        dic = {}
        for j in range(len(self.filter_names)):
            for i in range(len(self.user_prior.keys())):
                key = list(self.user_prior.keys())[i]
                dic[key + "_" + self.filter_names[j] + "_16"] = [quantiles[0, j, i]]
                dic[key + "_" + self.filter_names[j] + "_50"] = [quantiles[1, j, i]]
                dic[key + "_" + self.filter_names[j] + "_84"] = [quantiles[2, j, i]]

        df = pd.DataFrame(dic)
        df.to_csv(dir_path+'/mimical/cats' + f'/{self.id}.csv', index=False)


    

    def plot_model(self, type='median'):
        if self.success != True:
            print(f'Sampling failed, cannot plot model for {self.id}.')
        else:
            if type=='median':
                # Plot and save the median-parameter fit
                Plotter().plot_median(self.images, self.wavs, self.convolved_models, self.samples, list(self.fitter_prior.keys), self.prior_handler, self.filter_names)
                plt.savefig(dir_path+'/mimical/plots' + f'/{self.id}_median_model.pdf', bbox_inches='tight')
            elif type=='median-param':
                # Plot and save the median-parameter fit
                Plotter().plot_median_param(self.images, self.wavs, self.convolved_models, self.samples, list(self.fitter_prior.keys), self.prior_handler, self.filter_names)
                plt.savefig(dir_path+'/mimical/plots' + f'/{self.id}_median_param_model.pdf', bbox_inches='tight')

   



import matplotlib.pyplot as plt
import numpy as np
from astropy.convolution.utils import discretize_model
from tqdm import tqdm
from matplotlib import ticker

class Plotter(object):   

    def plot_median_param(self, images, wavs, convolved_models, samples, fitter_keys, prior_handler, filter_names):

        fig = plt.figure()
        gs = fig.add_gridspec(nrows=4, ncols=images.shape[0]+1, width_ratios=np.append(np.ones(images.shape[0]), 0.25))

        
        # Get median Nautilus parameters and transalte into median model parameters.
        param_dict = dict(zip(fitter_keys, np.median(samples, axis=0)))
        pars = prior_handler.revert(param_dict, wavs)


        models = np.zeros_like(images)
        for i in range(len(wavs)):
            convolved_models[i].parameters = pars[i]
            model = discretize_model(model=convolved_models[i], 
                                    x_range=[0,images[i].shape[1]], 
                                    y_range=[0,images[i].shape[0]], 
                                    mode='center')
            models[i]=model

        
        residuals = images - models

        vmins = [-np.percentile(images.flatten(), q=95), -np.percentile(images.flatten(), q=95), -np.percentile(images.flatten(), q=95), min(np.percentile(residuals.flatten(), q=5), -np.percentile(residuals.flatten(), q=95))]
        vmaxs = [np.percentile(images.flatten(), q=95), np.percentile(images.flatten(), q=95), np.percentile(images.flatten(), q=95), max(-np.percentile(residuals.flatten(), q=5), np.percentile(residuals.flatten(), q=95))]
        cmaps = ['binary', 'binary', 'RdGy']

        ax = fig.add_subplot(gs[0, 0])
        ax.set_axis_off()
        im1 = ax.pcolormesh(np.zeros_like(images[0]), vmax=vmaxs[0], vmin=vmins[0], cmap='RdGy')
        cbarax1 = fig.add_subplot(gs[:3, -1])
        cbarax1.set_yticks([])
        cbarax1.set_xticks([])
        cbar1 = plt.colorbar(im1, cax=cbarax1, fraction=1)
        tick_locator = ticker.MaxNLocator(nbins=5)
        cbar1.locator = tick_locator
        cbar1.update_ticks()

        im2 = ax.pcolormesh(np.zeros_like(images[0]), vmax=vmaxs[-1], vmin=vmins[-1], cmap='RdGy')
        cbarax2 = fig.add_subplot(gs[3, -1])
        cbarax2.set_yticks([])
        cbarax2.set_xticks([])
        cbar2 = plt.colorbar(im2, cax=cbarax2, fraction=1)
        tick_locator = ticker.MaxNLocator(nbins=3)
        cbar2.locator = tick_locator
        cbar2.update_ticks()

        for i in range(len(wavs)):

            plotims = [images[i], models[i], residuals[i], residuals[i]]

            for j in range(4):

                ax = fig.add_subplot(gs[j, i])
                im = ax.pcolormesh(plotims[j], vmax=vmaxs[j], vmin=vmins[j], cmap='RdGy')
                ax.set_yticks([])
                ax.set_xticks([])

                if j==0:
                    ax.set_title(filter_names[i].upper())

                if i==0:
                    if j==0:
                        ax.set_ylabel('Data')
                    if j==1:
                        ax.set_ylabel('Best\nModel')
                    if j==2:
                        ax.set_ylabel('Residual')
                    if j==3:
                        ax.set_ylabel('Residual\nZoom')

        
        plt.subplots_adjust(hspace=0.1, wspace=0.1)
        fig.set_size_inches(images.shape[0],4, forward=True)


    def plot_median(self, images, wavs, convolved_models, samples, fitter_keys, prior_handler, filter_names):

        fig = plt.figure()
        gs = fig.add_gridspec(nrows=4, ncols=images.shape[0]+1, width_ratios=np.append(np.ones(images.shape[0]), 0.25))
        
        models = np.zeros((samples.shape[0], *images.shape))

        print("Computing median model image...")
        for j in tqdm(range(samples.shape[0])):
            # Get median Nautilus parameters and transalte into median model parameters.
            param_dict = dict(zip(fitter_keys, samples[j]))
            pars = prior_handler.revert(param_dict, wavs)

            for k in range(len(wavs)):
                convolved_models[k].parameters = pars[k]
                model = discretize_model(model=convolved_models[k], 
                                        x_range=[0,images[k].shape[1]], 
                                        y_range=[0,images[k].shape[0]], 
                                        mode='center')
                models[j,k] = model


        models = np.median(models, axis=0)
        residuals = images - models


        vmins = [-np.percentile(images.flatten(), q=95), -np.percentile(images.flatten(), q=95), -np.percentile(images.flatten(), q=95), min(np.percentile(residuals.flatten(), q=5), -np.percentile(residuals.flatten(), q=95))]
        vmaxs = [np.percentile(images.flatten(), q=95), np.percentile(images.flatten(), q=95), np.percentile(images.flatten(), q=95), max(-np.percentile(residuals.flatten(), q=5), np.percentile(residuals.flatten(), q=95))]
        cmaps = ['binary', 'binary', 'RdGy']

        ax = fig.add_subplot(gs[0, 0])
        ax.set_axis_off()
        im1 = ax.pcolormesh(np.zeros_like(images[0]), vmax=vmaxs[0], vmin=vmins[0], cmap='RdGy')
        cbarax1 = fig.add_subplot(gs[:3, -1])
        cbarax1.set_yticks([])
        cbarax1.set_xticks([])
        cbar1 = plt.colorbar(im1, cax=cbarax1, fraction=1)
        tick_locator = ticker.MaxNLocator(nbins=5)
        cbar1.locator = tick_locator
        cbar1.update_ticks()

        im2 = ax.pcolormesh(np.zeros_like(images[0]), vmax=vmaxs[-1], vmin=vmins[-1], cmap='RdGy')
        cbarax2 = fig.add_subplot(gs[3, -1])
        cbarax2.set_yticks([])
        cbarax2.set_xticks([])
        cbar2 = plt.colorbar(im2, cax=cbarax2, fraction=1)
        tick_locator = ticker.MaxNLocator(nbins=3)
        cbar2.locator = tick_locator
        cbar2.update_ticks()

        for i in range(len(wavs)):

            plotims = [images[i], models[i], residuals[i], residuals[i]]

            for j in range(4):

                ax = fig.add_subplot(gs[j, i])
                im = ax.pcolormesh(plotims[j], vmax=vmaxs[j], vmin=vmins[j], cmap='RdGy')
                ax.set_yticks([])
                ax.set_xticks([])

                if j==0:
                    ax.set_title(filter_names[i].upper())

                if i==0:
                    if j==0:
                        ax.set_ylabel('Data')
                    if j==1:
                        ax.set_ylabel('Median\nModel')
                    if j==2:
                        ax.set_ylabel('Residual')
                    if j==3:
                        ax.set_ylabel('Residual\nZoom')

        
        plt.subplots_adjust(hspace=0.1, wspace=0.1)
        fig.set_size_inches(images.shape[0],4, forward=True)




import setuptools
from setuptools import setup
from os import path

here = path.abspath(path.dirname(__file__))
# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='mimical',

    version='0.1.5',

    description='Intesity modelling of multiply-imaged objects',

    long_description=long_description,

    author='Struan Stevenson',

    author_email='struan.stevenson@ed.ac.uk',

    packages= ["mimical", "mimical.fitting", "mimical.plotting", "mimical.utils",],

    include_package_data=True,

    install_requires=["numpy", "astropy", "matplotlib", "nautilus-sampler", "petrofit"],

)